/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { assert, unreachableCase } from "@fluidframework/common-utils";
// note: treats null and undefined as equivalent
export function areEqualAttributionKeys(a, b) {
    if (!a && !b) {
        return true;
    }
    if (!a || !b) {
        return false;
    }
    if (a.type !== b.type) {
        return false;
    }
    // Note: TS can't narrow the type of b inside this switch statement, hence the need for casting.
    switch (a.type) {
        case "op":
            return a.seq === b.seq;
        case "detached":
            return a.id === b.id;
        case "local":
            return true;
        default:
            unreachableCase(a, "Unhandled AttributionKey type");
    }
}
export class AttributionCollection {
    constructor(_length, baseEntry) {
        this._length = _length;
        this.offsets = [];
        this.keys = [];
        if (baseEntry !== undefined) {
            this.offsets.push(0);
            this.keys.push(baseEntry);
        }
    }
    get channelEntries() {
        var _a;
        return Object.entries((_a = this.channels) !== null && _a !== void 0 ? _a : {});
    }
    get channelNames() {
        var _a;
        return Object.keys((_a = this.channels) !== null && _a !== void 0 ? _a : {});
    }
    getAtOffset(offset, channel) {
        var _a;
        if (channel !== undefined) {
            const subCollection = (_a = this.channels) === null || _a === void 0 ? void 0 : _a[channel];
            return subCollection === null || subCollection === void 0 ? void 0 : subCollection.getAtOffset(offset);
        }
        assert(offset >= 0 && offset < this._length, 0x443 /* Requested offset should be valid */);
        return this.get(this.findIndex(offset));
    }
    findIndex(offset) {
        // Note: maximum length here is 256 for text segments. Perf testing shows that linear scan beats binary search
        // for attribution collections with under ~64 entries, and even at maximum size (which would require a maximum
        // length segment with every offset having different attribution), getAtOffset is on the order of 100ns.
        let i = 0;
        while (i < this.offsets.length && offset > this.offsets[i]) {
            i++;
        }
        return this.offsets[i] === offset ? i : i - 1;
    }
    get(index) {
        const key = this.keys[index];
        return key !== null ? key : undefined;
    }
    get length() {
        return this._length;
    }
    /**
     * Splits this attribution collection into two with entries for [0, pos) and [pos, length).
     */
    splitAt(pos) {
        const splitIndex = this.findIndex(pos);
        const splitCollection = new AttributionCollection(this.length - pos);
        for (let i = splitIndex; i < this.keys.length; i++) {
            splitCollection.offsets.push(Math.max(this.offsets[i] - pos, 0));
            splitCollection.keys.push(this.keys[i]);
        }
        if (this.channels) {
            splitCollection.channels = {};
            for (const [key, collection] of this.channelEntries) {
                splitCollection.channels[key] = collection.splitAt(pos);
            }
        }
        const spliceIndex = this.offsets[splitIndex] === pos ? splitIndex : splitIndex + 1;
        this.keys.splice(spliceIndex);
        this.offsets.splice(spliceIndex);
        this._length = pos;
        return splitCollection;
    }
    append(other) {
        var _a, _b, _c;
        var _d;
        const lastEntry = this.keys[this.keys.length - 1];
        for (let i = 0; i < other.keys.length; i++) {
            if (i !== 0 || !areEqualAttributionKeys(lastEntry, other.keys[i])) {
                this.offsets.push(other.offsets[i] + this.length);
                this.keys.push(other.keys[i]);
            }
        }
        if (other.channels !== undefined || this.channels !== undefined) {
            (_a = this.channels) !== null && _a !== void 0 ? _a : (this.channels = {});
            for (const [key, collection] of other.channelEntries) {
                const thisCollection = ((_b = (_d = this.channels)[key]) !== null && _b !== void 0 ? _b : (_d[key] = new AttributionCollection(this.length, null)));
                thisCollection.append(collection);
            }
            for (const [key, collection] of this.channelEntries) {
                if (((_c = other.channels) === null || _c === void 0 ? void 0 : _c[key]) === undefined) {
                    collection.append(new AttributionCollection(other.length, null));
                }
            }
        }
        this._length += other.length;
    }
    getAll() {
        const root = new Array(this.keys.length);
        for (let i = 0; i < this.keys.length; i++) {
            root[i] = { offset: this.offsets[i], key: this.keys[i] };
        }
        const result = {
            root,
            length: this.length,
        };
        if (this.channels !== undefined) {
            result.channels = {};
            for (const [key, collection] of this.channelEntries) {
                result.channels[key] = collection.getAll().root;
            }
        }
        return result;
    }
    clone() {
        const copy = new AttributionCollection(this.length);
        copy.keys = this.keys.slice();
        copy.offsets = this.offsets.slice();
        if (this.channels !== undefined) {
            const channelsCopy = {};
            for (const [key, collection] of this.channelEntries) {
                channelsCopy[key] = collection.clone();
            }
            copy.channels = channelsCopy;
        }
        return copy;
    }
    update(name, channel) {
        var _a;
        assert(channel.length === this.length, 0x5c0 /* AttributionCollection channel update should have consistent segment length */);
        if (name === undefined) {
            this.offsets = [...channel.offsets];
            this.keys = [...channel.keys];
        }
        else {
            (_a = this.channels) !== null && _a !== void 0 ? _a : (this.channels = {});
            if (this.channels[name] !== undefined) {
                this.channels[name].update(undefined, channel);
            }
            else {
                this.channels[name] = channel;
            }
        }
    }
    /**
     * Rehydrates attribution information from its serialized form into the provided iterable of consecutive segments.
     */
    static populateAttributionCollections(segments, summary) {
        const { channels } = summary;
        assert(summary.seqs.length === summary.posBreakpoints.length, 0x445 /* Invalid attribution summary blob provided */);
        const extractOntoSegments = ({ seqs, posBreakpoints }, assignToSegment) => {
            let curIndex = 0;
            let cumulativeSegPos = 0;
            for (const segment of segments) {
                const attribution = new AttributionCollection(segment.cachedLength);
                const pushEntry = (offset, seq) => {
                    attribution.offsets.push(offset);
                    attribution.keys.push(seq === null ? null : typeof seq === "object" ? seq : { type: "op", seq });
                };
                if (posBreakpoints[curIndex] > cumulativeSegPos) {
                    curIndex--;
                }
                while (posBreakpoints[curIndex] < cumulativeSegPos + segment.cachedLength) {
                    const nextOffset = Math.max(posBreakpoints[curIndex] - cumulativeSegPos, 0);
                    pushEntry(nextOffset, seqs[curIndex]);
                    curIndex++;
                }
                if (attribution.offsets.length === 0) {
                    pushEntry(0, seqs[curIndex - 1]);
                }
                assignToSegment(attribution, segment);
                cumulativeSegPos += segment.cachedLength;
            }
        };
        extractOntoSegments(summary, (collection, segment) => {
            segment.attribution = collection;
        });
        if (channels) {
            for (const [name, collectionSpec] of Object.entries(channels)) {
                extractOntoSegments(collectionSpec, (collection, segment) => {
                    var _a;
                    var _b;
                    // Cast is valid as we just assigned this field above
                    ((_a = (_b = segment.attribution).channels) !== null && _a !== void 0 ? _a : (_b.channels = {}))[name] =
                        collection;
                });
            }
        }
    }
    /**
     * Condenses attribution information on consecutive segments into a `SerializedAttributionCollection`
     *
     * Note: this operates on segments rather than attribution collections directly so that it can handle cases
     * where only some segments have attribution defined.
     */
    static serializeAttributionCollections(segments) {
        var _a;
        const allCollectionSpecs = [];
        const allChannelNames = new Set();
        for (const segment of segments) {
            const collection = (_a = segment.attribution) !== null && _a !== void 0 ? _a : new AttributionCollection(segment.cachedLength, null);
            const spec = collection.getAll();
            allCollectionSpecs.push(spec);
            if (spec.channels) {
                for (const name of Object.keys(spec.channels)) {
                    allChannelNames.add(name);
                }
            }
        }
        const extractSequenceOffsets = (getSpecEntries) => {
            const posBreakpoints = [];
            const seqs = [];
            let mostRecentAttributionKey;
            let cumulativePos = 0;
            for (const spec of allCollectionSpecs) {
                for (const { offset, key } of getSpecEntries(spec)) {
                    assert((key === null || key === void 0 ? void 0 : key.type) !== "local", 0x5c1 /* local attribution keys should never be put in summaries */);
                    if (mostRecentAttributionKey === undefined ||
                        !areEqualAttributionKeys(key, mostRecentAttributionKey)) {
                        posBreakpoints.push(offset + cumulativePos);
                        seqs.push(!key ? null : key.type === "op" ? key.seq : key);
                    }
                    mostRecentAttributionKey = key;
                }
                cumulativePos += spec.length;
            }
            return { seqs, posBreakpoints, length: cumulativePos };
        };
        const blobContents = extractSequenceOffsets((spec) => spec.root);
        if (allChannelNames.size > 0) {
            const channels = {};
            for (const name of allChannelNames) {
                const { posBreakpoints, seqs } = extractSequenceOffsets((spec) => { var _a, _b; return (_b = (_a = spec.channels) === null || _a === void 0 ? void 0 : _a[name]) !== null && _b !== void 0 ? _b : [{ offset: 0, key: null }]; });
                channels[name] = { posBreakpoints, seqs };
            }
            blobContents.channels = channels;
        }
        return blobContents;
    }
}
//# sourceMappingURL=attributionCollection.js.map